import sys
import os
import pytest
from liquidhandling import ExperimentManager
from liquidhandling import ExperimentParameter


class TestExperimentParameter:
    def test_temp(self):
        assert True
